from __future__ import annotations
from pathlib import Path

from .elements import Diagram, DiagramItem, Terminal
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (  # pragma: no cover
        Generator,
        Sequence as Seq,
        Tuple,
        TypeVar,
    )

    T = TypeVar("T")  # pragma: no cover
    Node = str | DiagramItem  # pragma: no cover


def write_diagram(diagram: Diagram, target: Path, standalone: bool = False):
    with open(target, "w") as t:
        if standalone:
            diagram.write_standalone(t.write)
        else:
            diagram.write_svg(t.write)


def escape_attr(val: str | float) -> str:
    if isinstance(val, str):
        return val.replace("&", "&amp;").replace("'", "&apos;").replace('"', "&quot;")
    return f"{val:g}"


def escape_html(val: str) -> str:
    return escape_attr(val).replace("<", "&lt;")


def determine_gaps(
    outer: float, inner: float, internal_alignment: str
) -> Tuple[float, float]:
    diff = outer - inner
    if internal_alignment == "left":
        return 0, diff
    elif internal_alignment == "right":
        return diff, 0
    else:
        return diff / 2, diff / 2


def double_enumerate(seq: Seq[T]) -> Generator[Tuple[int, int, T], None, None]:
    length = len(list(seq))
    for i, item in enumerate(seq):
        yield i, i - length, item


def add_debug(el: DiagramItem) -> None:
    if not el.parameters["debug"]:
        return
    el.attrs["data-x"] = "{0} w:{1} h:{2}/{3}/{4}".format(
        type(el).__name__, el.width, el.up, el.height, el.down
    )


def wrap_string(value: Node) -> DiagramItem:
    return value if isinstance(value, DiagramItem) else Terminal(value)



import typer

import json
import yaml

from typing_extensions import Annotated, Optional
from pathlib import Path

from .parser import parse_json, parse
from .utils import write_diagram

from rich.markdown import Markdown

cli = typer.Typer(add_completion=False, rich_markup_mode="rich", no_args_is_help=True)

input_file_argument = Annotated[
    Path,
    typer.Argument(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        show_default=False,
        help="Path to the input file.",
    ),
]

target_argument = Annotated[
    Path,
    typer.Argument(
        file_okay=True,
        dir_okay=False,
        writable=True,
        resolve_path=True,
        show_default=False,
        help="Path to the output file.",
    ),
]

parameters_argument = Annotated[
    Optional[Path],
    typer.Argument(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Path to the parameters file. If omitted, default parameters will be used.",
    ),
]

help_yaml = """
Parses a YAML [bold]FILE[/bold] for railroad diagrams and writes it into [bold]TARGET[/bold] file.
Various parameters of the diagram engine can be specified in a YAML [bold]PARAMETERS[/bold] file.

An input file would look like this:
[italic]
default: 0
element: Choice
items:
- element: Terminal
  text: foo
- element: Terminal
  href: raw
  text: bar
[/italic]
"""


@cli.command("yaml", no_args_is_help=True, help=help_yaml)
def parse_yaml_file(
    file: input_file_argument,
    target: target_argument,
    parameters: parameters_argument = None,
) -> None:
    if parameters:
        with open(parameters) as f:
            params = yaml.safe_load(f.read())
    else:
        params = {"standalone": False, "type": "complex", "css": None}
    with open(file) as f:
        yaml_input = yaml.safe_load(f.read())
        json_input = json.dumps(yaml_input)
        diagram = parse_json(json_input, params)
    if diagram:
        write_diagram(diagram, target, params["standalone"])


help_json = """
Parses a JSON [bold]FILE[/bold] for railroad diagrams and writes it into [bold]TARGET[/bold] file.
Various parameters of the diagram engine can be specified in a JSON [bold]PARAMETERS[/bold] file.

An input file would look like this:
[italic]
{
    "element": "Choice",
    "default": 0,
    "items": [
        {
            "element": "Terminal",
            "text": "foo"
        },
        {
            "element": "Terminal",
            "text": "bar",
            "href": "raw"
        }
    ]
}
[/italic]
"""


@cli.command("json", no_args_is_help=True, help=help_json)
def parse_json_file(
    file: input_file_argument,
    target: target_argument,
    parameters: parameters_argument = None,
) -> None:
    if parameters:
        with open(parameters, "r") as p:
            params = json.loads(p.read())
    else:
        params = {"standalone": False, "type": "complex", "css": None}
    with open(file) as f:
        diagram = parse_json(f.read(), params)
    if diagram:
        write_diagram(diagram, target, params["standalone"])


help_dsl = """
Parses a DSL [bold]FILE[/bold] for railroad diagrams, based on significant whitespace and writes it into [bold]TARGET[/bold] file.

Each command must be on its own line, and is written like "Sequence:\\n".

Children are indented on following lines.

Some commands have non-child arguments, for block commands they are put on the same line, after the :, for text commands they are put on the same line [italic]before[/italic] the :, like:
[italic]
Choice: 0
    Terminal: foo
    Terminal raw: bar
[/italic]
"""


@cli.command("dsl", no_args_is_help=True, help=help_dsl)
def parse_dsl_file(
    file: input_file_argument,
    target: target_argument,
    simple: Annotated[
        bool,
        typer.Option(
            "--simple",
            help="Toggle the simple diagram type. If omitted, uses the complex diagram type.",
        ),
    ] = False,
    standalone: Annotated[
        bool,
        typer.Option(
            "--standalone",
            help="Writes a standalone SVG containing the stylesheet to display it.",
        ),
    ] = False,
) -> None:
    with open(file) as f:
        diagram = parse(f.read(), simple)
    if diagram:
        write_diagram(diagram, target, standalone)


if __name__ == "__main__":
    cli()  # pragma: no cover



# -*- coding: utf-8 -*-
from __future__ import annotations

import math as Math

from typing import TYPE_CHECKING

from .exceptions import ParseException


if TYPE_CHECKING:
    from typing import (  # pragma: no cover
        Any,
        Callable,
        Dict,
        List,
        Optional as Opt,
        Sequence as Seq,
        TypeVar,
    )

    T = TypeVar("T")  # pragma: no cover
    Node = str | "DiagramItem"  # pragma: no cover
    WriterF = Callable[[str], Any]  # pragma: no cover
    WalkerF = Callable[["DiagramItem"], Any]  # pragma: no cover
    AttrsT = Dict[str, Any]  # pragma: no cover


class DiagramItem:
    def __init__(
        self,
        name: str,
        attrs: Opt[AttrsT] = None,
        text: Opt[Node] = None,
        parameters: Opt[AttrsT] = {},
    ):
        self.name = name
        # up = distance it projects above the entry line
        self.up: float = 0
        # height = distance between the entry/exit lines
        self.height: float = 0
        # down = distance it projects below the exit line
        self.down: float = 0
        # width = distance between the entry/exit lines horizontally
        self.width: float = 0
        # Whether the item is okay with being snug against another item or not
        self.needs_space = False

        # Parameters
        from .defaults import (
            DIAGRAM_CLASS,
            DEBUG,
            STROKE_ODD_PIXEL_LENGTH,
            VS,
            AR,
            CHAR_WIDTH,
            COMMENT_CHAR_WIDTH,
            INTERNAL_ALIGNMENT,
        )

        self.parameters = {
            "debug": DEBUG,
            "stroke_odd_pixel_length": STROKE_ODD_PIXEL_LENGTH,
            "diagram_class": DIAGRAM_CLASS,
            "VS": VS,
            "AR": AR,
            "char_width": CHAR_WIDTH,
            "comment_char_width": COMMENT_CHAR_WIDTH,
            "internal_alignment": INTERNAL_ALIGNMENT,
        }
        for k in parameters.keys():
            self.parameters[k] = parameters[k]

        # DiagramItems pull double duty as SVG elements.
        self.attrs: AttrsT = attrs or {}
        # Subclasses store their meaningful children as .item or .items;
        # .children instead stores their formatted SVG nodes.
        self.children: List[Node | Path | Style] = [text] if text else []

    def format(self, x: float, y: float, width: float) -> DiagramItem:
        raise NotImplementedError  # pragma: no cover

    def add_to(self, parent: DiagramItem) -> DiagramItem:
        parent.children.append(self)
        return self

    def write_svg(self, write: WriterF) -> None:
        from .utils import escape_attr, escape_html

        write("<{0}".format(self.name))
        for name, value in sorted(self.attrs.items()):
            write(' {0}="{1}"'.format(name, escape_attr(value)))
        write(">")
        if self.name in ["g", "svg"]:
            write("\n")
        for child in self.children:
            if isinstance(child, (DiagramItem, Path, Style)):
                child.write_svg(write)
            else:
                write(escape_html(child))
        write("</{0}>".format(self.name))

    def walk(self, cb: WalkerF) -> None:
        cb(self)

    def to_dict(self) -> dict:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def from_dict(
        cls, data: dict, parameters: dict | None = None
    ) -> DiagramItem | None:
        if "element" not in data.keys():
            return None
        match data["element"]:
            case "Diagram":
                return Diagram(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    type=parameters["type"],
                    parameters=parameters,
                )
            case "Start":
                if "type" not in data.keys() or data["type"] is None:
                    start_type = parameters.get("type", "simple")
                else:
                    start_type = data["type"]
                return Start(start_type, data.get("label", None), parameters=parameters)
            case "End":
                if "type" not in data.keys() or data["type"] is None:
                    end_type = parameters.get("type", "simple")
                else:
                    end_type = data["type"]
                return End(end_type, parameters=parameters)
            case "Terminal":
                return Terminal(
                    text=data["text"],
                    href=data.get("href", None),
                    title=data.get("title", None),
                    cls=data.get("cls", ""),
                    parameters=parameters,
                )
            case "NonTerminal":
                return NonTerminal(
                    text=data["text"],
                    href=data.get("href", None),
                    title=data.get("title", None),
                    cls=data.get("cls", ""),
                    parameters=parameters,
                )
            case "Stack":
                return Stack(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "Choice":
                try:
                    int(data["default"])
                except TypeError:
                    raise ParseException(
                        f"Attribute \"default\" must be an integer, got: {data['default']}."
                    )
                except KeyError:
                    data["default"] = 0
                return Choice(
                    int(data["default"]),
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "HorizontalChoice":
                return HorizontalChoice(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "OptionalSequence":
                return OptionalSequence(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "AlternatingSequence":
                return AlternatingSequence(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "MultipleChoice":
                return MultipleChoice(
                    int(data["default"]),
                    data["type"],
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "Skip":
                return Skip(parameters=parameters)
            case "OneOrMore":
                if "repeat" not in data.keys() or data["repeat"] is None:
                    return OneOrMore(
                        DiagramItem.from_dict(data["item"], parameters),
                        parameters=parameters,
                    )
                return OneOrMore(
                    DiagramItem.from_dict(data["item"], parameters),
                    DiagramItem.from_dict(data["repeat"], parameters),
                    parameters=parameters,
                )
            case "ZeroOrMore":
                if "repeat" not in data.keys() or data["repeat"] is None:
                    return zero_or_more(
                        DiagramItem.from_dict(data["item"], parameters),
                        skip=data.get("skip", False),
                        parameters=parameters,
                    )
                return zero_or_more(
                    DiagramItem.from_dict(data["item"], parameters),
                    DiagramItem.from_dict(data["repeat"], parameters),
                    skip=data.get("skip", False),
                    parameters=parameters,
                )
            case "Optional":
                return optional(
                    DiagramItem.from_dict(data["item"], parameters),
                    data.get("skip", False),
                    parameters=parameters,
                )
            case "Comment":
                return Comment(
                    text=data["text"],
                    href=data.get("href", None),
                    title=data.get("title", None),
                    cls=data.get("cls", ""),
                    parameters=parameters,
                )
            case "Sequence":
                return Sequence(
                    *(
                        DiagramItem.from_dict(item, parameters)
                        for item in data["items"]
                    ),
                    parameters=parameters,
                )
            case "Group":
                if "label" not in data.keys() or data["label"] is None:
                    return Group(
                        item=DiagramItem.from_dict(data["item"], parameters),
                        parameters=parameters,
                    )
                if isinstance(data["label"], str):
                    return Group(
                        item=DiagramItem.from_dict(data["item"], parameters),
                        label=data["label"],
                        parameters=parameters,
                    )
                return Group(
                    item=DiagramItem.from_dict(data["item"], parameters),
                    label=DiagramItem.from_dict(data["label"], parameters),
                    parameters=parameters,
                )
            case _:
                raise ParseException(f"Unknown element: {data['element']}.")


def apply_properties(properties: dict):
    """Need to make the global parameters not global"""
    pass


class DiagramMultiContainer(DiagramItem):
    def __init__(
        self,
        name: str,
        items: Seq[Node],
        attrs: Opt[Dict[str, str]] = None,
        text: Opt[str] = None,
        parameters: Opt[AttrsT] = {},
    ):
        DiagramItem.__init__(self, name, attrs, text, parameters=parameters)
        from .utils import wrap_string

        self.items: List[DiagramItem] = [wrap_string(item) for item in items]

    def format(self, x: float, y: float, width: float) -> DiagramItem:
        raise NotImplementedError  # Virtual

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        for item in self.items:
            item.walk(cb)


class Path:
    def __init__(self, x: float, y: float, cls: str = None, ar: float = None):
        self.x = x
        self.y = y
        self.AR = ar
        self.attrs = {"d": f"M{x} {y}"}
        if cls is not None:
            self.attrs = {"class": cls, "d": f"M{x} {y}"}

    def m(self, x: float, y: float) -> Path:
        self.attrs["d"] += f"m{x} {y}"
        return self

    def big_m(self, x: float, y: float) -> Path:
        self.attrs["d"] += f"M{x} {y}"
        return self

    def a(self, r: float) -> Path:
        self.attrs[
            "d"
        ] += f"a {r},{r} 0 0 1 -{r},{r} {r},{r} 0 0 1 -{r},-{r} {r},{r} 0 0 1 {r},-{r} {r},{r} 0 0 1 {r},{r} z"
        return self

    def l(self, x: float, y: float) -> Path:
        self.attrs["d"] += f"l{x} {y}"
        return self

    def h(self, val: float) -> Path:
        self.attrs["d"] += f"h{val}"
        return self

    def right(self, val: float) -> Path:
        return self.h(max(0, val))

    def left(self, val: float) -> Path:
        return self.h(-max(0, val))

    def v(self, val: float) -> Path:
        self.attrs["d"] += f"v{val}"
        return self

    def down(self, val: float) -> Path:
        return self.v(max(0, val))

    def up(self, val: float) -> Path:
        return self.v(-max(0, val))

    def arc_8(self, start: str, dir: str) -> Path:
        # 1/8 of a circle
        arc = self.AR
        s2 = 1 / Math.sqrt(2) * arc
        s2inv = arc - s2
        sweep = "1" if dir == "cw" else "0"
        path = f"a {arc} {arc} 0 0 {sweep} "
        sd = start + dir
        offset: List[float]
        match sd:
            case "ncw":
                offset = [s2, s2inv]
            case "necw":
                offset = [s2inv, s2]
            case "ecw":
                offset = [-s2inv, s2]
            case "secw":
                offset = [-s2, s2inv]
            case "scw":
                offset = [-s2, -s2inv]
            case "swcw":
                offset = [-s2inv, -s2]
            case "wcw":
                offset = [s2inv, -s2]
            case "nwcw":
                offset = [s2, -s2inv]
            case "nccw":
                offset = [-s2, s2inv]
            case "nwccw":
                offset = [-s2inv, s2]
            case "wccw":
                offset = [s2inv, s2]
            case "swccw":
                offset = [s2, s2inv]
            case "sccw":
                offset = [s2, -s2inv]
            case "seccw":
                offset = [s2inv, -s2]
            case "eccw":
                offset = [-s2inv, -s2]
            case "neccw":
                offset = [-s2, -s2inv]

        path += " ".join(str(x) for x in offset)
        self.attrs["d"] += path
        return self

    def arc(self, sweep: str) -> Path:
        x = self.AR
        y = self.AR
        if sweep[0] == "e" or sweep[1] == "w":
            x *= -1
        if sweep[0] == "s" or sweep[1] == "n":
            y *= -1
        cw = 1 if sweep in ("ne", "es", "sw", "wn") else 0
        self.attrs["d"] += f"a{self.AR} {self.AR} 0 0 {cw} {x} {y}"
        return self

    def add_to(self, parent: DiagramItem) -> Path:
        parent.children.append(self)
        return self

    def write_svg(self, write: WriterF) -> None:
        from .utils import escape_attr

        write("<path")
        for name, value in sorted(self.attrs.items()):
            write(f' {name}="{escape_attr(value)}"')
        write(" />")

    def format(self) -> Path:
        self.attrs["d"] += "h.5"
        return self

    def __repr__(self) -> str:
        return f"Path({repr(self.x)}, {repr(self.y)})"


class Style:
    def __init__(self, css: str):
        self.css = css

    def __repr__(self) -> str:
        return f"Style({repr(self.css)})"

    def add_to(self, parent: DiagramItem) -> Style:
        parent.children.append(self)
        return self

    def format(self) -> Style:
        return self

    def write_svg(self, write: WriterF) -> None:
        # Write included stylesheet as CDATA. See https:#developer.mozilla.org/en-US/docs/Web/SVG/Element/style
        cdata = "/* <![CDATA[ */\n{css}\n/* ]]> */\n".format(css=self.css)
        write("<style>{cdata}</style>".format(cdata=cdata))


class Diagram(DiagramMultiContainer):
    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}, **kwargs: str):
        # Accepts a type=[simple|complex] kwarg

        from .defaults import DIAGRAM_CLASS

        DiagramMultiContainer.__init__(
            self,
            "svg",
            list(items),
            {
                "class": parameters.get("diagram_class", DIAGRAM_CLASS),
            },
        )
        self.type = kwargs.get("type", "simple")
        if items and not isinstance(items[0], Start):
            self.items.insert(0, Start(self.type))
        if items and not isinstance(items[-1], End):
            self.items.append(End(self.type))
        self.up = 0
        self.down = 0
        self.height = 0
        self.width = 0
        for item in self.items:
            if isinstance(item, Style):
                continue
            self.width += item.width + (20 if item.needs_space else 0)
            self.up = max(self.up, item.up - self.height)
            self.height += item.height
            self.down = max(self.down - item.height, item.down)
        if self.items[0].needs_space:
            self.width -= 10
        if self.items[-1].needs_space:
            self.width -= 10
        self.formatted = False

    def to_dict(self) -> dict:
        return {"element": "Diagram", "items": [i.to_dict() for i in self.items]}

    def __repr__(self) -> str:
        items = ", ".join(map(repr, self.items[1:-1]))
        pieces = [] if not items else [items]
        if self.type != "simple":
            pieces.append(f"type={repr(self.type)}")
        return f'Diagram({", ".join(pieces)})'

    def format(
        self,
        padding_top: float = 20,
        padding_right: Opt[float] = None,
        padding_bottom: Opt[float] = None,
        padding_left: Opt[float] = None,
    ) -> Diagram:
        if padding_right is None:
            padding_right = padding_top
        if padding_bottom is None:
            padding_bottom = padding_top
        if padding_left is None:
            padding_left = padding_right
        assert padding_right is not None
        assert padding_bottom is not None
        assert padding_left is not None
        x = padding_left
        y = padding_top + self.up
        g = DiagramItem("g")
        if self.parameters["stroke_odd_pixel_length"]:
            g.attrs["transform"] = "translate(.5 .5)"
        for item in self.items:
            if item.needs_space:
                Path(x, y, ar=self.parameters["AR"]).h(10).add_to(g)
                x += 10
            item.format(x, y, item.width).add_to(g)
            x += item.width
            y += item.height
            if item.needs_space:
                Path(x, y, ar=self.parameters["AR"]).h(10).add_to(g)
                x += 10
        self.attrs["width"] = str(self.width + padding_left + padding_right)
        self.attrs["height"] = str(
            self.up + self.height + self.down + padding_top + padding_bottom
        )
        self.attrs["viewBox"] = f"0 0 {self.attrs['width']} {self.attrs['height']}"
        g.add_to(self)
        self.formatted = True
        return self

    def write_svg(self, write: WriterF) -> None:
        if not self.formatted:
            self.format()
        return DiagramItem.write_svg(self, write)

    def write_standalone(self, write: WriterF, css: str | None = None) -> None:
        if not self.formatted:
            self.format()
        if css is None:
            from importlib import resources as r
            from . import style

            inp_file = r.files(style) / "default.css"
            with inp_file.open("rt") as f:
                css = f.read()
        Style(css).add_to(self)
        self.attrs["xmlns"] = "http://www.w3.org/2000/svg"
        self.attrs["xmlns:xlink"] = "http://www.w3.org/1999/xlink"
        DiagramItem.write_svg(self, write)
        self.children.pop()
        del self.attrs["xmlns"]
        del self.attrs["xmlns:xlink"]


class Sequence(DiagramMultiContainer):
    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        self.needs_space = True
        self.up = 0
        self.down = 0
        self.height = 0
        self.width = 0
        for item in self.items:
            self.width += item.width + (20 if item.needs_space else 0)
            self.up = max(self.up, item.up - self.height)
            self.height += item.height
            self.down = max(self.down - item.height, item.down)
        if self.items[0].needs_space:
            self.width -= 10
        if self.items[-1].needs_space:
            self.width -= 10
        add_debug(self)

    def to_dict(self) -> dict:
        return {"element": "Sequence", "items": [i.to_dict() for i in self.items]}

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"Sequence({items})"

    def format(self, x: float, y: float, width: float) -> Sequence:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )
        Path(x, y, cls="seq seq1", ar=self.parameters["AR"]).h(left_gap).add_to(self)
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="seq seq2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap
        for i, item in enumerate(self.items):
            if item.needs_space and i > 0:
                Path(x, y, cls="seq seq3", ar=self.parameters["AR"]).h(10).add_to(self)
                x += 10
            item.format(x, y, item.width).add_to(self)
            x += item.width
            y += item.height
            if item.needs_space and i < len(self.items) - 1:
                Path(x, y, cls="seq seq4", ar=self.parameters["AR"]).h(10).add_to(self)
                x += 10
        return self


class Stack(DiagramMultiContainer):
    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        self.needs_space = True
        self.width = max(
            item.width + (20 if item.needs_space else 0) for item in self.items
        )
        # pretty sure that space calc is totes wrong
        if len(self.items) > 1:
            self.width += self.parameters["AR"] * 2
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = 0
        last = len(self.items) - 1
        for i, item in enumerate(self.items):
            self.height += item.height
            if i > 0:
                self.height += max(
                    self.parameters["AR"] * 2, item.up + self.parameters["VS"]
                )
            if i < last:
                self.height += max(
                    self.parameters["AR"] * 2, item.down + self.parameters["VS"]
                )
        add_debug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"Stack({items})"

    def to_dict(self) -> dict:
        return {"element": "Stack", "items": [i.to_dict() for i in self.items]}

    def format(self, x: float, y: float, width: float) -> Stack:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )
        Path(x, y, cls="stack stack1", ar=self.parameters["AR"]).h(left_gap).add_to(
            self
        )
        x += left_gap
        x_initial = x
        if len(self.items) > 1:
            Path(x, y, cls="stack stack2", ar=self.parameters["AR"]).h(
                self.parameters["AR"]
            ).add_to(self)
            x += self.parameters["AR"]
            inner_width = self.width - self.parameters["AR"] * 2
        else:
            inner_width = self.width
        for i, item in enumerate(self.items):
            item.format(x, y, inner_width).add_to(self)
            x += inner_width
            y += item.height
            if i != len(self.items) - 1:
                (
                    Path(x, y, cls="stack stack3", ar=self.parameters["AR"])
                    .arc("ne")
                    .down(
                        max(
                            0,
                            item.down
                            + self.parameters["VS"]
                            - self.parameters["AR"] * 2,
                        )
                    )
                    .arc("es")
                    .left(inner_width)
                    .arc("nw")
                    .down(
                        max(
                            0,
                            self.items[i + 1].up
                            + self.parameters["VS"]
                            - self.parameters["AR"] * 2,
                        )
                    )
                    .arc("ws")
                    .right(10)
                    .add_to(self)
                )
                y += max(
                    item.down + self.parameters["VS"], self.parameters["AR"] * 2
                ) + max(
                    self.items[i + 1].up + self.parameters["VS"],
                    self.parameters["AR"] * 2,
                )
                x = x_initial + self.parameters["AR"]
        if len(self.items) > 1:
            Path(x, y, cls="stack stack4", ar=self.parameters["AR"]).h(
                self.parameters["AR"]
            ).add_to(self)
            x += self.parameters["AR"]
        Path(x, y, cls="stack stack5", ar=self.parameters["AR"]).h(right_gap).add_to(
            self
        )
        return self


class OptionalSequence(DiagramMultiContainer):
    def __new__(cls, *items: Node, parameters: Opt[AttrsT] = {}) -> Any:
        if len(items) <= 1:
            return Sequence(*items, parameters=parameters)
        else:
            return super(OptionalSequence, cls).__new__(cls)

    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        self.needs_space = False
        self.width = 0
        self.up = 0
        self.height = sum(item.height for item in self.items)
        self.down = self.items[0].down
        height_so_far: float = 0
        for i, item in enumerate(self.items):
            self.up = max(
                self.up,
                max(self.parameters["AR"] * 2, item.up + self.parameters["VS"])
                - height_so_far,
            )
            height_so_far += item.height
            if i > 0:
                self.down = (
                    max(
                        self.height + self.down,
                        height_so_far
                        + max(
                            self.parameters["AR"] * 2, item.down + self.parameters["VS"]
                        ),
                    )
                    - self.height
                )
            item_width = item.width + (10 if item.needs_space else 0)
            if i == 0:
                self.width += self.parameters["AR"] + max(
                    item_width, self.parameters["AR"]
                )
            else:
                self.width += (
                    self.parameters["AR"] * 2
                    + max(item_width, self.parameters["AR"])
                    + self.parameters["AR"]
                )
        add_debug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"OptionalSequence({items})"

    def to_dict(self) -> dict:
        return {
            "element": "OptionalSequence",
            "items": [i.to_dict() for i in self.items],
        }

    def format(self, x: float, y: float, width: float) -> OptionalSequence:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )
        Path(x, y, cls="optseq os1", ar=self.parameters["AR"]).right(left_gap).add_to(
            self
        )
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="optseq os2",
            ar=self.parameters["AR"],
        ).right(right_gap).add_to(self)
        x += left_gap
        upper_line_y = y - self.up
        last = len(self.items) - 1
        for i, item in enumerate(self.items):
            item_space = 10 if item.needs_space else 0
            item_width = item.width + item_space
            if i == 0:
                # Upper skip
                (
                    Path(x, y, cls="optseq os3", ar=self.parameters["AR"])
                    .arc("se")
                    .up(y - upper_line_y - self.parameters["AR"] * 2)
                    .arc("wn")
                    .right(item_width - self.parameters["AR"])
                    .arc("ne")
                    .down(y + item.height - upper_line_y - self.parameters["AR"] * 2)
                    .arc("ws")
                    .add_to(self)
                )
                # Straight line
                (
                    Path(x, y, cls="optseq os4", ar=self.parameters["AR"])
                    .right(item_space + self.parameters["AR"])
                    .add_to(self)
                )
                item.format(
                    x + item_space + self.parameters["AR"], y, item.width
                ).add_to(self)
                x += item_width + self.parameters["AR"]
                y += item.height
            elif i < last:
                # Upper skip
                (
                    Path(x, upper_line_y, cls="optseq os5", ar=self.parameters["AR"])
                    .right(
                        self.parameters["AR"] * 2
                        + max(item_width, self.parameters["AR"])
                        + self.parameters["AR"]
                    )
                    .arc("ne")
                    .down(y - upper_line_y + item.height - self.parameters["AR"] * 2)
                    .arc("ws")
                    .add_to(self)
                )
                # Straight line
                (
                    Path(x, y, cls="optseq os6", ar=self.parameters["AR"])
                    .right(self.parameters["AR"] * 2)
                    .add_to(self)
                )
                item.format(x + self.parameters["AR"] * 2, y, item.width).add_to(self)
                (
                    Path(
                        x + item.width + self.parameters["AR"] * 2,
                        y + item.height,
                        cls="optseq os7",
                        ar=self.parameters["AR"],
                    )
                    .right(item_space + self.parameters["AR"])
                    .add_to(self)
                )
                # Lower skip
                (
                    Path(x, y, cls="optseq os8", ar=self.parameters["AR"])
                    .arc("ne")
                    .down(
                        item.height
                        + max(
                            item.down + self.parameters["VS"], self.parameters["AR"] * 2
                        )
                        - self.parameters["AR"] * 2
                    )
                    .arc("ws")
                    .right(item_width - self.parameters["AR"])
                    .arc("se")
                    .up(item.down + self.parameters["VS"] - self.parameters["AR"] * 2)
                    .arc("wn")
                    .add_to(self)
                )
                x += (
                    self.parameters["AR"] * 2
                    + max(item_width, self.parameters["AR"])
                    + self.parameters["AR"]
                )
                y += item.height
            else:
                # Straight line
                (
                    Path(x, y, cls="optseq os9", ar=self.parameters["AR"])
                    .right(self.parameters["AR"] * 2)
                    .add_to(self)
                )
                item.format(x + self.parameters["AR"] * 2, y, item.width).add_to(self)
                (
                    Path(
                        x + self.parameters["AR"] * 2 + item.width,
                        y + item.height,
                        cls="optseq os10",
                        ar=self.parameters["AR"],
                    )
                    .right(item_space + self.parameters["AR"])
                    .add_to(self)
                )
                # Lower skip
                (
                    Path(x, y, cls="optseq os11", ar=self.parameters["AR"])
                    .arc("ne")
                    .down(
                        item.height
                        + max(
                            item.down + self.parameters["VS"], self.parameters["AR"] * 2
                        )
                        - self.parameters["AR"] * 2
                    )
                    .arc("ws")
                    .right(item_width - self.parameters["AR"])
                    .arc("se")
                    .up(item.down + self.parameters["VS"] - self.parameters["AR"] * 2)
                    .arc("wn")
                    .add_to(self)
                )
        return self


class AlternatingSequence(DiagramMultiContainer):
    def __new__(cls, *items: Node, parameters: Opt[AttrsT] = {}) -> AlternatingSequence:
        if len(items) == 2:
            return super(AlternatingSequence, cls).__new__(cls)
        else:
            raise ParseException(
                "AlternatingSequence takes exactly two arguments, but got {0} arguments.".format(
                    len(items)
                )
            )

    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        self.needs_space = False

        arc = self.parameters["AR"]
        vert = self.parameters["VS"]
        first = self.items[0]
        second = self.items[1]

        arc_x = 1 / Math.sqrt(2) * arc * 2
        arc_y = (1 - 1 / Math.sqrt(2)) * arc * 2
        cross_y = max(arc, vert)
        cross_x = (cross_y - arc_y) + arc_x

        first_out = max(
            arc + arc, cross_y / 2 + arc + arc, cross_y / 2 + vert + first.down
        )
        self.up = first_out + first.height + first.up

        second_in = max(
            arc + arc, cross_y / 2 + arc + arc, cross_y / 2 + vert + second.up
        )
        self.down = second_in + second.height + second.down

        self.height = 0

        first_width = (20 if first.needs_space else 0) + first.width
        second_width = (20 if second.needs_space else 0) + second.width
        self.width = 2 * arc + max(first_width, cross_x, second_width) + 2 * arc
        add_debug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"AlternatingSequence({items})"

    def to_dict(self) -> dict:
        return {
            "element": "AlternatingSequence",
            "items": [i.to_dict() for i in self.items],
        }

    def format(self, x: float, y: float, width: float) -> AlternatingSequence:
        from .utils import determine_gaps

        arc = self.parameters["AR"]
        gaps = determine_gaps(width, self.width, self.parameters["internal_alignment"])
        Path(x, y, cls="altseq as1", ar=self.parameters["AR"]).right(gaps[0]).add_to(
            self
        )
        x += gaps[0]
        Path(x + self.width, y, cls="altseq as2", ar=self.parameters["AR"]).right(
            gaps[1]
        ).add_to(self)
        # bounding box
        # Path(x+gaps[0], y).up(self.up).right(self.width).down(self.up+self.down).left(self.width).up(self.down).addTo(self)
        first = self.items[0]
        second = self.items[1]

        # top
        first_in = self.up - first.up
        first_out = self.up - first.up - first.height
        Path(x, y, cls="altseq as3", ar=self.parameters["AR"]).arc("se").up(
            first_in - 2 * arc
        ).arc("wn").add_to(self)
        first.format(x + 2 * arc, y - first_in, self.width - 4 * arc).add_to(self)
        Path(
            x + self.width - 2 * arc,
            y - first_out,
            cls="altseq as4",
            ar=self.parameters["AR"],
        ).arc("ne").down(first_out - 2 * arc).arc("ws").add_to(self)

        # bottom
        second_in = self.down - second.down - second.height
        second_out = self.down - second.down
        Path(x, y, cls="altseq as5", ar=self.parameters["AR"]).arc("ne").down(
            second_in - 2 * arc
        ).arc("ws").add_to(self)
        second.format(x + 2 * arc, y + second_in, self.width - 4 * arc).add_to(self)
        Path(
            x + self.width - 2 * arc,
            y + second_out,
            cls="altseq as6",
            ar=self.parameters["AR"],
        ).arc("se").up(second_out - 2 * arc).arc("wn").add_to(self)

        # crossover
        arc_x = 1 / Math.sqrt(2) * arc * 2
        arc_y = (1 - 1 / Math.sqrt(2)) * arc * 2
        cross_y = max(arc, self.parameters["VS"])
        cross_x = (cross_y - arc_y) + arc_x
        cross_bar = (self.width - 4 * arc - cross_x) / 2
        (
            Path(
                x + arc,
                y - cross_y / 2 - arc,
                cls="altseq as7",
                ar=self.parameters["AR"],
            )
            .arc("ws")
            .right(cross_bar)
            .arc_8("n", "cw")
            .l(cross_x - arc_x, cross_y - arc_y)
            .arc_8("sw", "ccw")
            .right(cross_bar)
            .arc("ne")
            .add_to(self)
        )
        (
            Path(
                x + arc,
                y + cross_y / 2 + arc,
                cls="altseq as8",
                ar=self.parameters["AR"],
            )
            .arc("wn")
            .right(cross_bar)
            .arc_8("s", "ccw")
            .l(cross_x - arc_x, -(cross_y - arc_y))
            .arc_8("nw", "cw")
            .right(cross_bar)
            .arc("se")
            .add_to(self)
        )

        return self


class Choice(DiagramMultiContainer):
    def __init__(self, default: int, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        assert default < len(items)
        from .utils import add_debug

        self.default = default
        self.width = self.parameters["AR"] * 4 + max(item.width for item in self.items)
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = self.items[default].height
        for i, item in enumerate(self.items):
            if i in [default - 1, default + 1]:
                arcs = self.parameters["AR"] * 2
            else:
                arcs = self.parameters["AR"]
            if i < default:
                self.up += max(
                    arcs,
                    item.height
                    + item.down
                    + self.parameters["VS"]
                    + self.items[i + 1].up,
                )
            elif i == default:
                continue
            else:
                self.down += max(
                    arcs,
                    item.up
                    + self.parameters["VS"]
                    + self.items[i - 1].down
                    + self.items[i - 1].height,
                )
        self.down -= self.items[default].height  # already counted in self.height
        add_debug(self)

    def to_dict(self) -> dict:
        return {
            "element": "Choice",
            "default": self.default,
            "items": [i.to_dict() for i in self.items],
        }

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return "Choice(%r, %s)" % (self.default, items)

    def format(self, x: float, y: float, width: float) -> Choice:
        from .utils import determine_gaps, double_enumerate

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="choice ch1", ar=self.parameters["AR"]).h(left_gap).add_to(self)
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="choice ch2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap

        inner_width = self.width - self.parameters["AR"] * 4
        default = self.items[self.default]

        # Do the elements that curve above
        above = self.items[: self.default][::-1]
        if above:
            distance_from_y = max(
                self.parameters["AR"] * 2,
                default.up + self.parameters["VS"] + above[0].down + above[0].height,
            )
        for i, ni, item in double_enumerate(above):
            Path(x, y, cls="choice ch3", ar=self.parameters["AR"]).arc("se").up(
                distance_from_y - self.parameters["AR"] * 2
            ).arc("wn").add_to(self)
            item.format(
                x + self.parameters["AR"] * 2, y - distance_from_y, inner_width
            ).add_to(self)
            Path(
                x + self.parameters["AR"] * 2 + inner_width,
                y - distance_from_y + item.height,
                cls="choice ch4",
                ar=self.parameters["AR"],
            ).arc("ne").down(
                distance_from_y
                - item.height
                + default.height
                - self.parameters["AR"] * 2
            ).arc(
                "ws"
            ).add_to(
                self
            )
            if ni < -1:
                distance_from_y += max(
                    self.parameters["AR"],
                    item.up
                    + self.parameters["VS"]
                    + above[i + 1].down
                    + above[i + 1].height,
                )

        # Do the straight-line path.
        Path(x, y, cls="choice ch5", ar=self.parameters["AR"]).right(
            self.parameters["AR"] * 2
        ).add_to(self)
        self.items[self.default].format(
            x + self.parameters["AR"] * 2, y, inner_width
        ).add_to(self)
        Path(
            x + self.parameters["AR"] * 2 + inner_width,
            y + self.height,
            cls="choice ch6",
            ar=self.parameters["AR"],
        ).right(self.parameters["AR"] * 2).add_to(self)

        # Do the elements that curve below
        below = self.items[self.default + 1 :]
        if below:
            distance_from_y = max(
                self.parameters["AR"] * 2,
                default.height + default.down + self.parameters["VS"] + below[0].up,
            )
        for i, item in enumerate(below):
            Path(x, y, cls="choice ch7", ar=self.parameters["AR"]).arc("ne").down(
                distance_from_y - self.parameters["AR"] * 2
            ).arc("ws").add_to(self)
            item.format(
                x + self.parameters["AR"] * 2, y + distance_from_y, inner_width
            ).add_to(self)
            Path(
                x + self.parameters["AR"] * 2 + inner_width,
                y + distance_from_y + item.height,
                cls="choice ch8",
                ar=self.parameters["AR"],
            ).arc("se").up(
                distance_from_y
                - self.parameters["AR"] * 2
                + item.height
                - default.height
            ).arc(
                "wn"
            ).add_to(
                self
            )
            distance_from_y += max(
                self.parameters["AR"],
                item.height
                + item.down
                + self.parameters["VS"]
                + (below[i + 1].up if i + 1 < len(below) else 0),
            )
        return self


class MultipleChoice(DiagramMultiContainer):
    def __init__(
        self, default: int, type: str, *items: Node, parameters: Opt[AttrsT] = {}
    ):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        assert 0 <= default < len(items)
        assert type in ["any", "all"]
        self.default = default
        self.type = type
        self.needs_space = True
        self.inner_width = max(item.width for item in self.items)
        self.width = (
            30 + self.parameters["AR"] + self.inner_width + self.parameters["AR"] + 20
        )
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = self.items[default].height
        for i, item in enumerate(self.items):
            if i in [default - 1, default + 1]:
                minimum = 10 + self.parameters["AR"]
            else:
                minimum = self.parameters["AR"]
            if i < default:
                self.up += max(
                    minimum,
                    item.height
                    + item.down
                    + self.parameters["VS"]
                    + self.items[i + 1].up,
                )
            elif i == default:
                continue
            else:
                self.down += max(
                    minimum,
                    item.up
                    + self.parameters["VS"]
                    + self.items[i - 1].down
                    + self.items[i - 1].height,
                )
        self.down -= self.items[default].height  # already counted in self.height
        add_debug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"MultipleChoice({repr(self.default)}, {repr(self.type)}, {items})"

    def to_dict(self) -> dict:
        return {
            "element": "MultipleChoice",
            "default": self.default,
            "type": self.type,
            "items": [i.to_dict() for i in self.items],
        }

    def format(self, x: float, y: float, width: float) -> MultipleChoice:
        from .utils import determine_gaps, double_enumerate

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="multichoice mc1", ar=self.parameters["AR"]).h(left_gap).add_to(
            self
        )
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="multichoice mc2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap

        default = self.items[self.default]

        # Do the elements that curve above
        above = self.items[: self.default][::-1]
        if above:
            distance_from_y = max(
                10 + self.parameters["AR"],
                default.up + self.parameters["VS"] + above[0].down + above[0].height,
            )
        for i, ni, item in double_enumerate(above):
            (
                Path(x + 30, y, cls="multichoice mc3", ar=self.parameters["AR"])
                .up(distance_from_y - self.parameters["AR"])
                .arc("wn")
                .add_to(self)
            )
            item.format(
                x + 30 + self.parameters["AR"], y - distance_from_y, self.inner_width
            ).add_to(self)
            (
                Path(
                    x + 30 + self.parameters["AR"] + self.inner_width,
                    y - distance_from_y + item.height,
                    cls="multichoice mc4",
                    ar=self.parameters["AR"],
                )
                .arc("ne")
                .down(
                    distance_from_y
                    - item.height
                    + default.height
                    - self.parameters["AR"]
                    - 10
                )
                .add_to(self)
            )
            if ni < -1:
                distance_from_y += max(
                    self.parameters["AR"],
                    item.up
                    + self.parameters["VS"]
                    + above[i + 1].down
                    + above[i + 1].height,
                )

        # Do the straight-line path.
        Path(x + 30, y, cls="multichoice mc5", ar=self.parameters["AR"]).right(
            self.parameters["AR"]
        ).add_to(self)
        self.items[self.default].format(
            x + 30 + self.parameters["AR"], y, self.inner_width
        ).add_to(self)
        Path(
            x + 30 + self.parameters["AR"] + self.inner_width,
            y + self.height,
            cls="multichoice mc6",
            ar=self.parameters["AR"],
        ).right(self.parameters["AR"]).add_to(self)

        # Do the elements that curve below
        below = self.items[self.default + 1 :]
        if below:
            distance_from_y = max(
                10 + self.parameters["AR"],
                default.height + default.down + self.parameters["VS"] + below[0].up,
            )
        for i, item in enumerate(below):
            (
                Path(x + 30, y, cls="multichoice mc7", ar=self.parameters["AR"])
                .down(distance_from_y - self.parameters["AR"])
                .arc("ws")
                .add_to(self)
            )
            item.format(
                x + 30 + self.parameters["AR"], y + distance_from_y, self.inner_width
            ).add_to(self)
            (
                Path(
                    x + 30 + self.parameters["AR"] + self.inner_width,
                    y + distance_from_y + item.height,
                    cls="multichoice mc8",
                    ar=self.parameters["AR"],
                )
                .arc("se")
                .up(
                    distance_from_y
                    - self.parameters["AR"]
                    + item.height
                    - default.height
                    - 10
                )
                .add_to(self)
            )
            distance_from_y += max(
                self.parameters["AR"],
                item.height
                + item.down
                + self.parameters["VS"]
                + (below[i + 1].up if i + 1 < len(below) else 0),
            )
        text = DiagramItem("g", attrs={"class": "diagram-text"}).add_to(self)
        DiagramItem(
            "title",
            text="take one or more branches, once each, in any order"
            if self.type == "any"
            else "take all branches, once each, in any order",
        ).add_to(text)
        DiagramItem(
            "path",
            attrs={
                "d": "M {x} {y} h -26 a 4 4 0 0 0 -4 4 v 12 a 4 4 0 0 0 4 4 h 26 z".format(
                    x=x + 30, y=y - 10
                ),
                "class": "diagram-text",
            },
        ).add_to(text)
        DiagramItem(
            "text",
            text="1+" if self.type == "any" else "all",
            attrs={"x": x + 15, "y": y + 4, "class": "diagram-text"},
        ).add_to(text)
        DiagramItem(
            "path",
            attrs={
                "d": "M {x} {y} h 16 a 4 4 0 0 1 4 4 v 12 a 4 4 0 0 1 -4 4 h -16 z".format(
                    x=x + self.width - 20, y=y - 10
                ),
                "class": "diagram-text",
            },
        ).add_to(text)
        DiagramItem(
            "text",
            text="↺",
            attrs={"x": x + self.width - 10, "y": y + 4, "class": "diagram-arrow"},
        ).add_to(text)
        return self


class HorizontalChoice(DiagramMultiContainer):
    def __new__(cls, *items: Node, parameters: Opt[AttrsT] = {}) -> Any:
        if len(items) <= 1:
            return Sequence(*items, parameters=parameters)
        else:
            return super(HorizontalChoice, cls).__new__(cls)

    def __init__(self, *items: Node, parameters: Opt[AttrsT] = {}):
        DiagramMultiContainer.__init__(self, "g", items, parameters=parameters)
        from .utils import add_debug

        all_but_last = self.items[:-1]
        middles = self.items[1:-1]
        first = self.items[0]
        last = self.items[-1]
        self.needs_space = False

        self.width = (
            self.parameters["AR"]  # starting track
            + self.parameters["AR"] * 2 * (len(self.items) - 1)  # in-between tracks
            + sum(x.width + (20 if x.needs_space else 0) for x in self.items)  # items
            + (
                self.parameters["AR"] if last.height > 0 else 0
            )  # needs space to curve up
            + self.parameters["AR"]
        )  # ending track

        # Always exits at entrance height
        self.height = 0

        # All but the last have a track running above them
        self._upperTrack = max(
            self.parameters["AR"] * 2,
            self.parameters["VS"],
            max(x.up for x in all_but_last) + self.parameters["VS"],
        )
        self.up = max(self._upperTrack, last.up)

        # All but the first have a track running below them
        # Last either straight-lines or curves up, so has different calculation
        self._lowerTrack = max(
            self.parameters["VS"],
            max(
                x.height
                + max(x.down + self.parameters["VS"], self.parameters["AR"] * 2)
                for x in middles
            )
            if middles
            else 0,
            last.height + last.down + self.parameters["VS"],
        )
        if first.height < self._lowerTrack:
            # Make sure there's at least 2*self.parameters["AR"] room between first exit and lower track
            self._lowerTrack = max(
                self._lowerTrack, first.height + self.parameters["AR"] * 2
            )
        self.down = max(self._lowerTrack, first.height + first.down)

        add_debug(self)

    def to_dict(self) -> dict:
        return {
            "element": "HorizontalChoice",
            "items": [i.to_dict() for i in self.items],
        }

    def format(self, x: float, y: float, width: float) -> HorizontalChoice:
        from .utils import determine_gaps

        # Hook up the two sides if self is narrower than its stated width.
        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )
        Path(x, y, cls="horizchoice hc1", ar=self.parameters["AR"]).h(left_gap).add_to(
            self
        )
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="horizchoice hc2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap

        first = self.items[0]
        last = self.items[-1]

        # upper track
        upper_span = (
            sum(x.width + (20 if x.needs_space else 0) for x in self.items[:-1])
            + (len(self.items) - 2) * self.parameters["AR"] * 2
            - self.parameters["AR"]
        )
        (
            Path(x, y, cls="horizchoice hc3", ar=self.parameters["AR"])
            .arc("se")
            .up(self._upperTrack - self.parameters["AR"] * 2)
            .arc("wn")
            .h(upper_span)
            .add_to(self)
        )

        # lower track
        lower_span = (
            sum(x.width + (20 if x.needs_space else 0) for x in self.items[1:])
            + (len(self.items) - 2) * self.parameters["AR"] * 2
            + (self.parameters["AR"] if last.height > 0 else 0)
            - self.parameters["AR"]
        )
        lower_start = (
            x
            + self.parameters["AR"]
            + first.width
            + (20 if first.needs_space else 0)
            + self.parameters["AR"] * 2
        )
        (
            Path(
                lower_start,
                y + self._lowerTrack,
                cls="horizchoice hc4",
                ar=self.parameters["AR"],
            )
            .h(lower_span)
            .arc("se")
            .up(self._lowerTrack - self.parameters["AR"] * 2)
            .arc("wn")
            .add_to(self)
        )

        # Items
        for [i, item] in enumerate(self.items):
            # input track
            if i == 0:
                (
                    Path(x, y, cls="horizchoice hc5", ar=self.parameters["AR"])
                    .h(self.parameters["AR"])
                    .add_to(self)
                )
                x += self.parameters["AR"]
            else:
                (
                    Path(
                        x,
                        y - self._upperTrack,
                        cls="horizchoice hc6",
                        ar=self.parameters["AR"],
                    )
                    .arc("ne")
                    .v(self._upperTrack - self.parameters["AR"] * 2)
                    .arc("ws")
                    .add_to(self)
                )
                x += self.parameters["AR"] * 2

            # item
            item_width = item.width + (20 if item.needs_space else 0)
            item.format(x, y, item_width).add_to(self)
            x += item_width

            # output track
            if i == len(self.items) - 1:
                if item.height == 0:
                    (
                        Path(x, y, cls="horizchoice hc7", ar=self.parameters["AR"])
                        .h(self.parameters["AR"])
                        .add_to(self)
                    )
                else:
                    (
                        Path(
                            x,
                            y + item.height,
                            cls="horizchoice hc8",
                            ar=self.parameters["AR"],
                        )
                        .arc("se")
                        .add_to(self)
                    )
            elif i == 0 and item.height > self._lowerTrack:
                # Needs to arc up to meet the lower track, not down.
                if item.height - self._lowerTrack >= self.parameters["AR"] * 2:
                    (
                        Path(
                            x,
                            y + item.height,
                            cls="horizchoice hc9",
                            ar=self.parameters["AR"],
                        )
                        .arc("se")
                        .v(self._lowerTrack - item.height + self.parameters["AR"] * 2)
                        .arc("wn")
                        .add_to(self)
                    )
                else:
                    # Not enough space to fit two arcs
                    # so just bail and draw a straight line for now.
                    (
                        Path(
                            x,
                            y + item.height,
                            cls="horizchoice hc10",
                            ar=self.parameters["AR"],
                        )
                        .l(self.parameters["AR"] * 2, self._lowerTrack - item.height)
                        .add_to(self)
                    )
            else:
                (
                    Path(
                        x,
                        y + item.height,
                        cls="horizchoice hc11",
                        ar=self.parameters["AR"],
                    )
                    .arc("ne")
                    .v(self._lowerTrack - item.height - self.parameters["AR"] * 2)
                    .arc("ws")
                    .add_to(self)
                )
        return self


def optional(item: Node, skip: bool = False, parameters: Opt[AttrsT] = {}) -> Choice:
    return Choice(
        0 if skip else 1, Skip(parameters=parameters), item, parameters=parameters
    )


class OneOrMore(DiagramItem):
    def __init__(
        self, item: Node, repeat: Opt[Node] = None, parameters: Opt[AttrsT] = {}
    ):
        DiagramItem.__init__(self, "g", parameters=parameters)
        from .utils import add_debug, wrap_string

        self.item = wrap_string(item)
        repeat = repeat or Skip()
        self.rep = wrap_string(repeat)
        self.width = max(self.item.width, self.rep.width) + self.parameters["AR"] * 2
        self.height = self.item.height
        self.up = self.item.up
        self.down = max(
            self.parameters["AR"] * 2,
            self.item.down
            + self.parameters["VS"]
            + self.rep.up
            + self.rep.height
            + self.rep.down,
        )
        self.needs_space = True
        add_debug(self)

    def format(self, x: float, y: float, width: float) -> OneOrMore:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="oneor oom1", ar=self.parameters["AR"]).h(left_gap).add_to(self)
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="oneor oom2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap

        # Draw item
        Path(x, y, cls="oneor oom3", ar=self.parameters["AR"]).right(
            self.parameters["AR"]
        ).add_to(self)
        self.item.format(
            x + self.parameters["AR"], y, self.width - self.parameters["AR"] * 2
        ).add_to(self)
        Path(
            x + self.width - self.parameters["AR"],
            y + self.height,
            cls="oneor oom4",
            ar=self.parameters["AR"],
        ).right(self.parameters["AR"]).add_to(self)

        # Draw repeat arc
        distance_from_y = max(
            self.parameters["AR"] * 2,
            self.item.height + self.item.down + self.parameters["VS"] + self.rep.up,
        )
        Path(
            x + self.parameters["AR"], y, cls="oneor oom5", ar=self.parameters["AR"]
        ).arc("nw").down(distance_from_y - self.parameters["AR"] * 2).arc("ws").add_to(
            self
        )
        self.rep.format(
            x + self.parameters["AR"],
            y + distance_from_y,
            self.width - self.parameters["AR"] * 2,
        ).add_to(self)
        Path(
            x + self.width - self.parameters["AR"],
            y + distance_from_y + self.rep.height,
            cls="oneor oom6",
            ar=self.parameters["AR"],
        ).arc("se").up(
            distance_from_y
            - self.parameters["AR"] * 2
            + self.rep.height
            - self.item.height
        ).arc(
            "en"
        ).add_to(
            self
        )

        return self

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        self.item.walk(cb)
        self.rep.walk(cb)

    def __repr__(self) -> str:
        return f"OneOrMore({repr(self.item)}, repeat={repr(self.rep)})"

    def to_dict(self) -> dict:
        return {
            "element": "OneOrMore",
            "item": self.item.to_dict(),
            "repeat": self.rep.to_dict(),
        }


def zero_or_more(
    item: Node,
    repeat: Opt[Node] = None,
    skip: bool = False,
    parameters: Opt[AttrsT] = {},
) -> Choice:
    result = optional(
        OneOrMore(item, repeat, parameters=parameters), skip, parameters=parameters
    )
    return result


class Group(DiagramItem):
    def __init__(
        self, item: Node, label: Opt[Node] = None, parameters: Opt[AttrsT] = {}
    ):
        DiagramItem.__init__(self, "g", parameters=parameters)
        from .utils import add_debug, wrap_string

        self.item = wrap_string(item)
        self.label: Opt[DiagramItem]
        if isinstance(label, DiagramItem):
            self.label = label
        elif label:
            self.label = Comment(label)
        else:
            self.label = None

        self.width = max(
            self.item.width + (20 if self.item.needs_space else 0),
            self.label.width if self.label else 0,
            self.parameters["AR"] * 2,
        )
        self.height = self.item.height
        self.boxUp = max(self.item.up + self.parameters["VS"], self.parameters["AR"])
        self.up = self.boxUp
        if self.label:
            self.up += self.label.up + self.label.height + self.label.down
        self.down = max(self.item.down + self.parameters["VS"], self.parameters["AR"])
        self.needs_space = True
        add_debug(self)

    def format(self, x: float, y: float, width: float) -> Group:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )
        Path(x, y, cls="group gr1", ar=self.parameters["AR"]).h(left_gap).add_to(self)
        Path(
            x + left_gap + self.width,
            y + self.height,
            cls="group gr2",
            ar=self.parameters["AR"],
        ).h(right_gap).add_to(self)
        x += left_gap

        DiagramItem(
            "rect",
            {
                "x": x,
                "y": y - self.boxUp,
                "width": self.width,
                "height": self.boxUp + self.height + self.down,
                "rx": self.parameters["AR"],
                "ry": self.parameters["AR"],
                "class": "group-box",
            },
        ).add_to(self)

        self.item.format(x, y, self.width).add_to(self)
        if self.label:
            self.label.format(
                x,
                y - (self.boxUp + self.label.down + self.label.height),
                self.label.width,
            ).add_to(self)

        return self

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        self.item.walk(cb)
        if self.label:
            self.label.walk(cb)

    def to_dict(self) -> dict:
        if self.label is None:
            return {"element": "Group", "item": self.item.to_dict(), "label": None}
        return {
            "element": "Group",
            "item": self.item.to_dict(),
            "label": self.label.to_dict(),
        }


class Start(DiagramItem):
    def __init__(
        self, type: str = "simple", label: Opt[str] = None, parameters: Opt[AttrsT] = {}
    ):
        DiagramItem.__init__(self, "g", parameters=parameters)
        from .utils import add_debug

        if label:
            self.width = max(20, len(label) * self.parameters["char_width"] + 10)
        else:
            self.width = 20
        self.up = 10
        self.down = 10
        self.type = type
        self.label = label
        add_debug(self)

    def format(self, x: float, y: float, width: float) -> Start:
        path = Path(x, y - 10, cls="start")
        if self.type == "complex":
            path.down(20).m(0, -10).right(self.width).add_to(self)
        elif self.type == "sql":
            path = Path(x, y - 10, cls="start ")
            path.m(0, 10).a(3.7).big_m(x, y).right(self.width).add_to(self)
        else:
            path.down(20).m(10, -20).down(20).m(-10, -10).right(self.width).add_to(self)
        if self.label:
            DiagramItem(
                "text",
                attrs={"x": x, "y": y - 15, "style": "text-anchor:start"},
                text=self.label,
            ).add_to(self)
        return self

    def __repr__(self) -> str:
        return f"Start(type={repr(self.type)}, label={repr(self.label)})"

    def to_dict(self) -> dict:
        return {"element": "Start", "type": self.type, "label": self.label}


class End(DiagramItem):
    def __init__(self, type: str = "simple", parameters: Opt[AttrsT] = {}):
        DiagramItem.__init__(self, "path", parameters=parameters)
        from .utils import add_debug

        self.width = 20
        self.up = 10
        self.down = 10
        self.type = type
        add_debug(self)

    def format(self, x: float, y: float, width: float) -> End:
        self.attrs["class"] = "end"
        if self.type == "simple":
            self.attrs["d"] = "M {0} {1} h 20 m -10 -10 v 20 m 10 -20 v 20".format(x, y)
        elif self.type == "complex":
            self.attrs["d"] = "M {0} {1} h 20 m 0 -10 v 20".format(x, y)
        elif self.type == "sql":
            self.attrs["d"] = "M {0} {1} h 20 m -5 -5 5,5 -5,5".format(x, y)
        return self

    def __repr__(self) -> str:
        return f"End(type={repr(self.type)})"

    def to_dict(self) -> dict:
        return {"element": "End", "type": self.type}


class Terminal(DiagramItem):
    def __init__(
        self,
        text: str,
        href: Opt[str] = None,
        title: Opt[str] = None,
        cls: str = "",
        parameters: Opt[AttrsT] = {},
    ):
        DiagramItem.__init__(
            self, "g", {"class": " ".join(["terminal", cls])}, parameters=parameters
        )
        from .utils import add_debug

        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * self.parameters["char_width"] + 20
        self.up = 11
        self.down = 11
        self.needs_space = True
        add_debug(self)

    def to_dict(self) -> dict:
        return {
            "element": "Terminal",
            "text": self.text,
            "href": self.href,
            "title": self.title,
            "cls": self.cls,
        }

    def __repr__(self) -> str:
        return f"Terminal({repr(self.text)}, href={repr(self.href)}, title={repr(self.title)}, cls={repr(self.cls)})"

    def format(self, x: float, y: float, width: float) -> Terminal:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="terminal term1", ar=self.parameters["AR"]).h(left_gap).add_to(
            self
        )
        Path(
            x + left_gap + self.width, y, cls="terminal term2", ar=self.parameters["AR"]
        ).h(right_gap).add_to(self)

        DiagramItem(
            "rect",
            {
                "x": x + left_gap,
                "y": y - 11,
                "width": self.width,
                "height": self.up + self.down,
                "rx": 10,
                "ry": 10,
            },
        ).add_to(self)
        text = DiagramItem(
            "text", {"x": x + left_gap + self.width / 2, "y": y + 4}, self.text
        )
        if self.href is not None:
            DiagramItem("a", {"xlink:href": self.href}, text).add_to(self)
        else:
            text.add_to(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).add_to(self)
        return self


class NonTerminal(DiagramItem):
    def __init__(
        self,
        text: str,
        href: Opt[str] = None,
        title: Opt[str] = None,
        cls: str = "",
        parameters: Opt[AttrsT] = {},
    ):
        DiagramItem.__init__(
            self, "g", {"class": " ".join(["non-terminal", cls])}, parameters=parameters
        )
        from .utils import add_debug

        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * self.parameters["char_width"] + 20
        self.up = 11
        self.down = 11
        self.needs_space = True
        add_debug(self)

    def to_dict(self) -> dict:
        return {
            "element": "NonTerminal",
            "text": self.text,
            "href": self.href,
            "title": self.title,
            "cls": self.cls,
        }

    def __repr__(self) -> str:
        return f"NonTerminal({repr(self.text)}, href={repr(self.href)}, title={repr(self.title)}, cls={repr(self.cls)})"

    def format(self, x: float, y: float, width: float) -> NonTerminal:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="nonterm nt1", ar=self.parameters["AR"]).h(left_gap).add_to(self)
        Path(
            x + left_gap + self.width, y, cls="nonterm nt2", ar=self.parameters["AR"]
        ).h(right_gap).add_to(self)

        DiagramItem(
            "rect",
            {
                "x": x + left_gap,
                "y": y - 11,
                "width": self.width,
                "height": self.up + self.down,
            },
        ).add_to(self)
        text = DiagramItem(
            "text", {"x": x + left_gap + self.width / 2, "y": y + 4}, self.text
        )
        if self.href is not None:
            DiagramItem("a", {"xlink:href": self.href}, text).add_to(self)
        else:
            text.add_to(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).add_to(self)
        return self


class Comment(DiagramItem):
    def __init__(
        self,
        text: str,
        href: Opt[str] = None,
        title: Opt[str] = None,
        cls: str = "",
        parameters: Opt[AttrsT] = {},
    ):
        DiagramItem.__init__(
            self, "g", {"class": " ".join(["non-terminal", cls])}, parameters=parameters
        )
        from .utils import add_debug

        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * self.parameters["comment_char_width"] + 10
        self.up = 8
        self.down = 8
        self.needs_space = True
        add_debug(self)

    def to_dict(self) -> dict:
        return {
            "element": "Comment",
            "text": self.text,
            "href": self.href,
            "title": self.title,
            "cls": self.cls,
        }

    def __repr__(self) -> str:
        return f"Comment({repr(self.text)}, href={repr(self.href)}, title={repr(self.title)}, cls={repr(self.cls)})"

    def format(self, x: float, y: float, width: float) -> Comment:
        from .utils import determine_gaps

        left_gap, right_gap = determine_gaps(
            width, self.width, self.parameters["internal_alignment"]
        )

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y, cls="comment com1", ar=self.parameters["AR"]).h(left_gap).add_to(
            self
        )
        Path(
            x + left_gap + self.width, y, cls="comment com2", ar=self.parameters["AR"]
        ).h(right_gap).add_to(self)

        text = DiagramItem(
            "text",
            {"x": x + left_gap + self.width / 2, "y": y + 5, "class": "comment"},
            self.text,
        )
        if self.href is not None:
            DiagramItem("a", {"xlink:href": self.href}, text).add_to(self)
        else:
            text.add_to(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).add_to(self)
        return self


class Skip(DiagramItem):
    def __init__(self, parameters: Opt[AttrsT] = {}) -> None:
        DiagramItem.__init__(self, "g", parameters=parameters)
        from .utils import add_debug

        self.width = 0
        self.up = 0
        self.down = 0
        add_debug(self)

    def format(self, x: float, y: float, width: float) -> Skip:
        Path(x, y, cls="skip", ar=self.parameters["AR"]).right(width).add_to(self)
        return self

    def __repr__(self) -> str:
        return "Skip()"

    def to_dict(self) -> dict:
        return {"element": "Skip"}



import os
import errno
import unittest
import pytest


def silent_remove(filename):
    try:
        os.remove(filename)
    except OSError as e:  # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred


class BaseTest(unittest.TestCase):
    def setUp(self) -> None:
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()


class UnitTests(BaseTest):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_terminal(self):
        from pyrailroad.elements import Terminal, Diagram

        with pytest.raises(TypeError):
            Terminal()
        t = Terminal("text")
        assert t.to_dict() == {
            "element": "Terminal",
            "text": "text",
            "href": None,
            "title": None,
            "cls": "",
        }
        t = Terminal("text", "href")
        assert t.to_dict() == {
            "element": "Terminal",
            "text": "text",
            "href": "href",
            "title": None,
            "cls": "",
        }
        t = Terminal("text", "href", "title")
        assert t.to_dict() == {
            "element": "Terminal",
            "text": "text",
            "href": "href",
            "title": "title",
            "cls": "",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/terminal.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/terminal_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_non_terminal(self):
        from pyrailroad.elements import NonTerminal, Diagram

        with pytest.raises(TypeError):
            NonTerminal()
        t = NonTerminal("text")
        assert t.to_dict() == {
            "element": "NonTerminal",
            "text": "text",
            "href": None,
            "title": None,
            "cls": "",
        }
        t = NonTerminal("text", "href")
        assert t.to_dict() == {
            "element": "NonTerminal",
            "text": "text",
            "href": "href",
            "title": None,
            "cls": "",
        }
        t = NonTerminal("text", "href", "title")
        assert t.to_dict() == {
            "element": "NonTerminal",
            "text": "text",
            "href": "href",
            "title": "title",
            "cls": "",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/non_terminal.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/non_terminal_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_comment(self):
        from pyrailroad.elements import Comment, Diagram

        with pytest.raises(TypeError):
            Comment()
        t = Comment("text")
        assert t.to_dict() == {
            "element": "Comment",
            "text": "text",
            "href": None,
            "title": None,
            "cls": "",
        }
        t = Comment("text", "href")
        assert t.to_dict() == {
            "element": "Comment",
            "text": "text",
            "href": "href",
            "title": None,
            "cls": "",
        }
        t = Comment("text", "href", "title")
        assert t.to_dict() == {
            "element": "Comment",
            "text": "text",
            "href": "href",
            "title": "title",
            "cls": "",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/comment.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/comment_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_skip(self):
        from pyrailroad.elements import Skip, Diagram

        t = Skip()
        assert t.to_dict() == {"element": "Skip"}
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/skip.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/skip_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_sequence(self):
        from pyrailroad.elements import Terminal, Sequence, Diagram

        t = Sequence(Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "element": "Sequence",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/sequence.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/sequence_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_stack(self):
        from pyrailroad.elements import Terminal, Diagram, Stack

        t = Stack(Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "element": "Stack",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/stack.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/stack_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_optional_sequence(self):
        from pyrailroad.elements import Terminal, OptionalSequence, Diagram

        t = OptionalSequence(Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "element": "OptionalSequence",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/optional_sequence.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/optional_sequence_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_alternating_sequence(self):
        from pyrailroad.elements import Terminal, AlternatingSequence, Diagram

        t = AlternatingSequence(Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "element": "AlternatingSequence",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/alternating_sequence.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/alternating_sequence_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_choice(self):
        from pyrailroad.elements import Terminal, Choice, Diagram

        with pytest.raises(TypeError):
            Choice(Terminal("term1"), Terminal("term2"))
        t = Choice(0, Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 0,
            "element": "Choice",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/choice0.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/choice0_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        t = Choice(1, Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 1,
            "element": "Choice",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/choice1.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/choice1_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_multiple_choice(self):
        from pyrailroad.elements import Terminal, MultipleChoice, Diagram

        with pytest.raises(TypeError):
            MultipleChoice()
        with pytest.raises(TypeError):
            MultipleChoice(Terminal("term1"))
        with pytest.raises(TypeError):
            MultipleChoice(Terminal("term1"), Terminal("term2"))
        t = MultipleChoice(0, "all", Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 0,
            "element": "MultipleChoice",
            "type": "all",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/multiple_choice0_all.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/multiple_choice0_all_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        t = MultipleChoice(1, "all", Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 1,
            "element": "MultipleChoice",
            "type": "all",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/multiple_choice1_all.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/multiple_choice1_all_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        t = MultipleChoice(0, "any", Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 0,
            "element": "MultipleChoice",
            "type": "any",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/multiple_choice0_any.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/multiple_choice0_any_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        t = MultipleChoice(1, "any", Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "default": 1,
            "element": "MultipleChoice",
            "type": "any",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/multiple_choice1_any.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/multiple_choice1_any_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_horizontal_choice(self):
        from pyrailroad.elements import Terminal, HorizontalChoice, Diagram

        with pytest.raises(IndexError):
            HorizontalChoice()

        t = HorizontalChoice(Terminal("term1"))
        assert t.to_dict() == {
            "element": "Sequence",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                }
            ],
        }

        t = HorizontalChoice(Terminal("term1"), Terminal("term2"))
        assert t.to_dict() == {
            "element": "HorizontalChoice",
            "items": [
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term1",
                    "title": None,
                },
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term2",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/horizontal_choice.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/horizontal_choice_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_optional(self):
        from pyrailroad.elements import Terminal, optional, Diagram

        t = optional(Terminal("term"))
        assert t.to_dict() == {
            "default": 1,
            "element": "Choice",
            "items": [
                {"element": "Skip"},
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/optional_no_skip.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/optional_no_skip_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = optional(Terminal("term"), True)
        assert t.to_dict() == {
            "default": 0,
            "element": "Choice",
            "items": [
                {"element": "Skip"},
                {
                    "cls": "",
                    "element": "Terminal",
                    "href": None,
                    "text": "term",
                    "title": None,
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/optional_skip.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/optional_skip_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_one_or_more(self):
        from pyrailroad.elements import Diagram, OneOrMore, Terminal

        with pytest.raises(TypeError):
            OneOrMore()

        t = OneOrMore(Terminal("term"))
        assert t.to_dict() == {
            "element": "OneOrMore",
            "item": {
                "cls": "",
                "element": "Terminal",
                "href": None,
                "text": "term",
                "title": None,
            },
            "repeat": {"element": "Skip"},
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/one_or_more_skip.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/one_or_more_skip_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = OneOrMore(Terminal("term"), Terminal("repeat"))
        assert t.to_dict() == {
            "element": "OneOrMore",
            "item": {
                "cls": "",
                "element": "Terminal",
                "href": None,
                "text": "term",
                "title": None,
            },
            "repeat": {
                "cls": "",
                "element": "Terminal",
                "href": None,
                "text": "repeat",
                "title": None,
            },
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/one_or_more_repeat.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/one_or_more_repeat_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_zero_or_more(self):
        from pyrailroad.elements import Diagram, zero_or_more, Terminal

        with pytest.raises(TypeError):
            zero_or_more()  # NOSONAR

        t = zero_or_more(Terminal("term"))
        assert t.to_dict() == {
            "element": "Choice",
            "default": 1,
            "items": [
                {"element": "Skip"},
                {
                    "element": "OneOrMore",
                    "item": {
                        "element": "Terminal",
                        "text": "term",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                    "repeat": {"element": "Skip"},
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/zero_or_more_skip1.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/zero_or_more_skip1_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = zero_or_more(Terminal("term"), skip=True)
        assert t.to_dict() == {
            "element": "Choice",
            "default": 0,
            "items": [
                {"element": "Skip"},
                {
                    "element": "OneOrMore",
                    "item": {
                        "element": "Terminal",
                        "text": "term",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                    "repeat": {"element": "Skip"},
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/zero_or_more_skip0.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/zero_or_more_skip0_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = zero_or_more(Terminal("term"), Terminal("repeat"))
        assert t.to_dict() == {
            "element": "Choice",
            "default": 1,
            "items": [
                {"element": "Skip"},
                {
                    "element": "OneOrMore",
                    "item": {
                        "element": "Terminal",
                        "text": "term",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                    "repeat": {
                        "element": "Terminal",
                        "text": "repeat",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/zero_or_more_repeat1.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/zero_or_more_repeat1_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = zero_or_more(Terminal("term"), Terminal("repeat"), skip=True)
        assert t.to_dict() == {
            "element": "Choice",
            "default": 0,
            "items": [
                {"element": "Skip"},
                {
                    "element": "OneOrMore",
                    "item": {
                        "element": "Terminal",
                        "text": "term",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                    "repeat": {
                        "element": "Terminal",
                        "text": "repeat",
                        "href": None,
                        "title": None,
                        "cls": "",
                    },
                },
            ],
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/zero_or_more_repeat0.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/zero_or_more_repeat0_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_group(self):
        from pyrailroad.elements import Diagram, Group, Terminal

        with pytest.raises(TypeError):
            Group()

        t = Group(Terminal("term"))
        assert t.to_dict() == {
            "element": "Group",
            "item": {
                "element": "Terminal",
                "text": "term",
                "href": None,
                "title": None,
                "cls": "",
            },
            "label": None,
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/group_no_label.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/group_no_label_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = Group(Terminal("term"), "label")
        assert t.to_dict() == {
            "element": "Group",
            "item": {
                "element": "Terminal",
                "text": "term",
                "href": None,
                "title": None,
                "cls": "",
            },
            "label": {
                "element": "Comment",
                "text": "label",
                "href": None,
                "title": None,
                "cls": "",
            },
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/group_label.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/group_label_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_start(self):
        from pyrailroad.elements import Start, Diagram

        t = Start()
        assert t.to_dict() == {"element": "Start", "type": "simple", "label": None}
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/start_simple.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/start_simple_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = Start(label="label")
        assert t.to_dict() == {"element": "Start", "type": "simple", "label": "label"}
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/start_label.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/start_label_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = Start("complex")
        assert t.to_dict() == {"element": "Start", "type": "complex", "label": None}
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/start_complex.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/start_complex_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = Start("sql")
        assert t.to_dict() == {"element": "Start", "type": "sql", "label": None}
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/start_sql.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/start_sql_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

    def test_end(self):
        from pyrailroad.elements import End, Diagram

        t = End()
        assert t.to_dict() == {
            "element": "End",
            "type": "simple",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/end_simple.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/end_simple_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = End("complex")
        assert t.to_dict() == {
            "element": "End",
            "type": "complex",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/end_complex.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/end_complex_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result

        t = End("sql")
        assert t.to_dict() == {
            "element": "End",
            "type": "sql",
        }
        d = Diagram(t)
        svg = []
        d.write_svg(svg.append)
        with open("tests/end_sql.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result
        svg = []
        d.write_standalone(svg.append)
        with open("tests/end_sql_standalone.svg", "r") as f:
            svg_result = f.read()
        assert " ".join(svg) == svg_result


class JSONParserTests(BaseTest):
    def test_parse_json(self):
        from pyrailroad.parser import parse_json
        from pyrailroad.elements import Diagram

        input_string = """{
    "element": "Choice",
    "default": 0,
    "items": [
        {
            "element": "Terminal",
            "text": "foo"
        },
        {
            "element": "Terminal",
            "text": "bar",
            "href": "raw"
        }
    ]
}"""
        r = parse_json(
            input_string, {"standalone": False, "type": "complex", "css": None}
        )
        assert isinstance(r, Diagram)

    def test_missing_element_json(self):
        from pyrailroad.parser import parse_json
        from pyrailroad.exceptions import ParseException

        input_string = """{
    "error": "Choice",
    "default": 0,
    "items": [
        {
            "element": "Terminal",
            "text": "foo"
        },
        {
            "element": "Terminal",
            "text": "bar",
            "href": "raw"
        }
    ]
}"""
        with pytest.raises(ParseException) as e:
            parse_json(
                input_string, {"standalone": False, "type": "complex", "css": None}
            )
        assert e.value.msg == "Invalid input file : 'element' is missing from the root."

    def test_choice_error_json(self):
        from pyrailroad.parser import parse_json
        from pyrailroad.exceptions import ParseException

        input_string = """{
    "element": "Choice",
    "default": null,
    "items": [
        {
            "element": "Terminal",
            "text": "foo"
        },
        {
            "element": "Terminal",
            "text": "bar",
            "href": "raw"
        }
    ]
}"""
        with pytest.raises(ParseException) as e:
            parse_json(
                input_string, {"standalone": False, "type": "complex", "css": None}
            )
        assert e.value.msg == 'Attribute "default" must be an integer, got: None.'

    def test_unknown_element_json(self):
        from pyrailroad.parser import parse_json
        from pyrailroad.exceptions import ParseException

        input_string = """{
    "element": "Chance",
    "default": 0,
    "items": [
        {
            "element": "Terminal",
            "text": "foo"
        },
        {
            "element": "Terminal",
            "text": "bar",
            "href": "raw"
        }
    ]
}"""
        with pytest.raises(ParseException) as e:
            parse_json(
                input_string, {"standalone": False, "type": "complex", "css": None}
            )
        assert e.value.msg == "Unknown element: Chance."


class DSLExceptionTests(BaseTest):
    def test_general_parsing_errors(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        unknown_element = "Terminally: foo"
        with pytest.raises(ParseException) as e:
            parse(unknown_element, True)
        assert (
            e.value.msg
            == "Line 1 doesn't contain a valid railroad-diagram command. Got:\nTerminally: foo"
        )

        bad_indent = "    Choice: 0\n  Terminal: foo\n  Terminal: bar"
        with pytest.raises(ParseException) as e:
            parse(bad_indent, True)
        assert (
            e.value.msg
            == "Inconsistent indentation: line 1 is indented less than the first line."
        )

        bad_block_grammar = "Sequence foo:\n  Terminal: bar\n  Terminal: bar"
        with pytest.raises(ParseException) as e:
            parse(bad_block_grammar, True)
        assert (
            e.value.msg
            == "Line 1 doesn't match the grammar 'Command: optional-prelude'. Got:\nSequence foo:"
        )

        bad_multiple_choice_grammar = "MultipleChoice foo: bar"
        with pytest.raises(ParseException) as e:
            parse(bad_multiple_choice_grammar, True)
        assert (
            e.value.msg
            == "Line 1 doesn't match the grammar 'MultipleChoice: optional-prelude'. Got:\nMultipleChoice foo: bar"
        )

        bad_indents = "Sequence:\n\tTerminal: foo\n\tTerminal: bar\nSequence:\n\t\tTerminal: foo\n\t\tTerminal: bar"
        with pytest.raises(ParseException) as e:
            parse(bad_indents, True)
        assert (
            e.value.msg
            == "Line 5 jumps more than 1 indent level from the previous line:\nTerminal: foo"
        )

    def test_terminal_parsing_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Terminal:\n  Terminal: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Terminal commands cannot have children."

    def test_non_terminal_parsing_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "NonTerminal:\n  Terminal: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - NonTerminal commands cannot have children."

    def test_comment_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Comment:\n  Terminal: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Comment commands cannot have children."

    def test_skip_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Skip:\n  Terminal: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Skip commands cannot have children."

        element = "Skip: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Skip commands cannot have text."

    def test_sequence_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Sequence: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Sequence commands cannot have preludes."

        element = "Sequence:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Sequence commands need at least one child."

    def test_stack_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Stack: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Stack commands cannot have preludes."

        element = "Stack:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Stack commands need at least one child."

    def test_horizontal_choice_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "HorizontalChoice: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - HorizontalChoice commands cannot have preludes."

        element = "HorizontalChoice:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - HorizontalChoice commands need at least one child."
        )

    def test_optional_sequence_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "OptionalSequence: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - OptionalSequence commands cannot have preludes."

        element = "OptionalSequence:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - OptionalSequence commands need at least one child."
        )

    def test_alternating_sequence_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "AlternatingSequence: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - AlternatingSequence commands cannot have preludes."
        )

        element = "AlternatingSequence:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg
            == "Line 1 - AlternatingSequence commands need exactly two children."
        )

    def test_one_or_more_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "OneOrMore: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - OneOrMore commands cannot have preludes."

        element = "OneOrMore:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - OneOrMore commands must have one or two children."
        )

    def test_zero_or_more_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "ZeroOrMore: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg
            == "Line 1 - ZeroOrMore preludes must be nothing or 'skip'. Got:\nfoo"
        )

        element = "ZeroOrMore:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - ZeroOrMore commands must have one or two children."
        )

    def test_group_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Group:\n\tTerminal: foo\n\tTerminal: bar"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Group commands need exactly one child."

    def test_optional_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Optional: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg
            == "Line 1 - Optional preludes must be nothing or 'skip'. Got:\nfoo"
        )

        element = "Optional:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Optional commands need exactly one child."

    def test_choice_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "Choice: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Choice preludes must be an integer. Got:\nfoo"

        element = "Choice:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - Choice commands need at least one child."

    def test_multiple_choice_error(self):
        from pyrailroad.parser import parse
        from pyrailroad.exceptions import ParseException

        element = "MultipleChoice: foo"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert e.value.msg == "Line 1 - MultipleChoice type must be any or all."

        element = "MultipleChoice:"
        with pytest.raises(ParseException) as e:
            parse(element, True)
        assert (
            e.value.msg == "Line 1 - MultipleChoice commands need at least one child."
        )


class CLITests(BaseTest):
    def setUp(self):
        super().setUp()
        from typer.testing import CliRunner

        self.runner = CliRunner()

    def tearDown(self):
        silent_remove("tests/cli/output.svg")
        super().tearDown()

    def test_cli_help(self):
        from pyrailroad.cli import cli

        result = self.runner.invoke(cli, ["--help"])
        assert result.exit_code == 0
        assert "dsl" in result.stdout
        assert "json" in result.stdout
        assert "yaml" in result.stdout
        result = self.runner.invoke(cli, ["dsl", "--help"])
        assert result.exit_code == 0
        assert "file" in result.stdout
        assert "target" in result.stdout
        result = self.runner.invoke(cli, ["json", "--help"])
        assert result.exit_code == 0
        assert "file" in result.stdout
        assert "target" in result.stdout
        assert "parameters" in result.stdout
        result = self.runner.invoke(cli, ["yaml", "--help"])
        assert result.exit_code == 0
        assert "file" in result.stdout
        assert "target" in result.stdout
        assert "parameters" in result.stdout

    def test_cli_dsl(self):
        from pyrailroad.cli import cli

        in_file = "tests/cli/diagram.dsl"
        out_file = "tests/cli/output.svg"
        result = self.runner.invoke(cli, ["dsl", in_file, out_file])
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(cli, ["dsl", in_file, out_file, "--standalone"])
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["dsl", in_file, out_file, "--standalone", "--simple"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone_simple.svg", "r") as base:
                assert res.read() == base.read()

    def test_cli_json(self):
        from pyrailroad.cli import cli

        in_file = "tests/cli/diagram.json"
        out_file = "tests/cli/output.svg"
        result = self.runner.invoke(cli, ["json", in_file, out_file])
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["json", in_file, out_file, "tests/cli/complex_standalone.json"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["json", in_file, out_file, "tests/cli/simple_standalone.json"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone_simple.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["json", in_file, out_file, "tests/cli/customized_standalone.json"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone_custom.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["yaml", in_file, out_file, "tests/cli/sql_standalone.json"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_sql_standalone.svg", "r") as base:
                assert res.read() == base.read()

    def test_cli_yaml(self):
        from pyrailroad.cli import cli

        in_file = "tests/cli/diagram.yaml"
        out_file = "tests/cli/output.svg"
        result = self.runner.invoke(cli, ["yaml", in_file, out_file])
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["yaml", in_file, out_file, "tests/cli/complex_standalone.yaml"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["yaml", in_file, out_file, "tests/cli/simple_standalone.yaml"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone_simple.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["yaml", in_file, out_file, "tests/cli/customized_standalone.yaml"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_standalone_custom.svg", "r") as base:
                assert res.read() == base.read()

        result = self.runner.invoke(
            cli, ["yaml", in_file, out_file, "tests/cli/sql_standalone.yaml"]
        )
        assert result.exit_code == 0
        with open(out_file, "r") as res:
            with open("tests/cli/diagram_sql_standalone.svg", "r") as base:
                assert res.read() == base.read()



import dataclasses
import json
import typing as t
from .elements import (
    Diagram,
    Terminal,
    Comment,
    NonTerminal,
    Skip,
    Sequence,
    Stack,
    MultipleChoice,
    Choice,
    HorizontalChoice,
    AlternatingSequence,
    OptionalSequence,
    DiagramItem,
    OneOrMore,
    Group,
    zero_or_more,
    optional,
)

from .exceptions import ParseException


@dataclasses.dataclass
class RRCommand:
    name: str
    prelude: str | None
    children: list["RRCommand"]
    text: str | None
    line: int


def parse_json(string: str, properties: {}) -> Diagram | None:
    data = json.loads(string)
    if "element" not in data:
        raise ParseException("Invalid input file : 'element' is missing from the root.")
    if data["element"] != "Diagram":
        input_data = {"element": "Diagram", "items": [data]}
    else:
        input_data = data
    diagram = Diagram.from_dict(input_data, properties)
    return diagram


def parse(string: str, simple: bool) -> Diagram | None:
    """
    Parses a DSL for railroad diagrams, based on significant whitespace.
    Each command must be on its own line, and is written like "Sequence:\n".
    Children are indented on following lines.
    Some commands have non-child arguments;
    for block commands they are put on the same line, after the :,
    for text commands they are put on the same line *before* the :,
    like:
        Choice: 0
            Terminal: foo
            Terminal raw: bar
    """
    import re

    diagram_type = "complex"
    if simple:
        diagram_type = "simple"

    lines = string.splitlines()

    # Strip off any common initial whitespace from lines.
    initial_indent = t.cast(re.Match, re.match(r"(\s*)", lines[0])).group(1)
    for i, line in enumerate(lines):
        if line.startswith(initial_indent):
            lines[i] = line[len(initial_indent) :]
        else:
            raise ParseException(
                f"Inconsistent indentation: line {i} is indented less than the first line."
            )

    # Determine subsequent indentation
    for line in lines:
        match = re.match(r"(\s+)", line)
        if match:
            indent_text = match.group(1)
            break
    else:
        indent_text = "\t"

    # Turn lines into tree
    last_indent = 0
    tree = RRCommand(name="Diagram", prelude="", children=[], text=None, line=0)
    active_commands = {"0": tree}
    block_names = "And|Seq|Sequence|Stack|Or|Choice|Opt|Optional|Plus|OneOrMore|Star|ZeroOrMore|OptionalSequence|HorizontalChoice|AlternatingSequence|Group"
    text_names = "T|Terminal|N|NonTerminal|C|Comment|S|Skip"
    for i, line in enumerate(lines, 1):
        indent = 0
        while line.startswith(indent_text):
            indent += 1
            line = line[len(indent_text) :]
        if indent > last_indent + 1:
            raise ParseException(
                f"Line {i} jumps more than 1 indent level from the previous line:\n{line.strip()}"
            )
        last_indent = indent
        if re.match(rf"\s*({block_names})\W", line):
            match = re.match(r"\s*(\w+)\s*:\s*(.*)", line)
            if not match:
                raise ParseException(
                    f"Line {i} doesn't match the grammar 'Command: optional-prelude'. Got:\n{line.strip()}"
                )
            command = match.group(1)
            prelude = match.group(2).strip()
            node = RRCommand(
                name=command, prelude=prelude, children=[], text=None, line=i
            )
        elif re.match(r"\s*(MultipleChoice)\W", line):
            match = re.match(r"\s*(\w+)\s*:\s*(\d*)\s*(.*)", line)
            if not match:
                raise ParseException(
                    f"Line {i} doesn't match the grammar 'MultipleChoice: optional-prelude'. Got:\n{line.strip()}"
                )
            command = match.group(1)
            prelude = [match.group(2).strip(), match.group(3).strip()]
            node = RRCommand(
                name=command, prelude=prelude, children=[], text=None, line=i
            )
        elif re.match(rf"\s*({text_names})\W", line):
            match = re.match(r"\s*(\w+)\s*(\"[\w\s/:.-]+\"|[\w\s]+)?:\s*(.*)", line)
            if not match:  # Unreachable?
                raise ParseException(
                    f"Line {i} doesn't match the grammar 'Command [optional prelude]: text'. Got:\n{line.strip()},"
                )
            command = match.group(1)
            if match.group(2):
                prelude = match.group(2).strip().strip('"')
            else:
                prelude = None
            text = match.group(3).strip()
            node = RRCommand(
                name=command,
                prelude=prelude,
                children=[],
                text=text,
                line=i,
            )
        else:
            raise ParseException(
                f"Line {i} doesn't contain a valid railroad-diagram command. Got:\n{line.strip()}",
            )

        active_commands[str(indent)].children.append(node)
        active_commands[str(indent + 1)] = node

    diagram = create_diagram(tree, diagram_type=diagram_type)
    assert diagram is None or isinstance(diagram, Diagram)
    return diagram


def create_diagram_node(command: RRCommand, diagram_type: str) -> Diagram:
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return Diagram(*children, type=diagram_type)


def create_terminal_node(command: RRCommand) -> Terminal | None:
    if command.children:
        raise ParseException(
            f"Line {command.line} - Terminal commands cannot have children."
        )
    return Terminal(command.text or "", command.prelude)


def create_non_terminal_node(command: RRCommand) -> NonTerminal | None:
    if command.children:
        raise ParseException(
            f"Line {command.line} - NonTerminal commands cannot have children."
        )
    return NonTerminal(command.text or "", command.prelude)


def create_comment_node(command: RRCommand) -> Comment | None:
    if command.children:
        raise ParseException(
            f"Line {command.line} - Comment commands cannot have children."
        )
    return Comment(command.text or "", command.prelude)


def create_skip_node(command: RRCommand) -> Skip | None:
    if command.children:
        raise ParseException(
            f"Line {command.line} - Skip commands cannot have children."
        )
    if command.text:
        raise ParseException(f"Line {command.line} - Skip commands cannot have text.")
    return Skip()


def create_sequence_node(command: RRCommand) -> Sequence | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - Sequence commands cannot have preludes."
        )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - Sequence commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return Sequence(*children)


def create_stack_node(command: RRCommand) -> Stack | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - Stack commands cannot have preludes."
        )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - Stack commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return Stack(*children)


def create_horizontal_choice_node(command: RRCommand) -> HorizontalChoice | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - HorizontalChoice commands cannot have preludes."
        )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - HorizontalChoice commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return HorizontalChoice(*children)


def create_multiple_choice_node(command: RRCommand) -> MultipleChoice | None:
    if (default := command.prelude[0]) == "":
        default = 0
    try:
        default = int(t.cast(str, default))
    except ValueError:
        raise ParseException(  # Unreachable?
            f"Line {command.line} - MultipleChoice preludes must be an integer. Got:\n{command.prelude}"
        )
    if (mc_type := command.prelude[1]) == "":
        mc_type = "any"
    if mc_type not in ["any", "all"]:
        raise ParseException(
            f"Line {command.line} - MultipleChoice type must be any or all."
        )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - MultipleChoice commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return MultipleChoice(default, mc_type, *children)


def create_optional_sequence_node(command: RRCommand) -> OptionalSequence | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - OptionalSequence commands cannot have preludes."
        )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - OptionalSequence commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return OptionalSequence(*children)


def create_choice_node(command: RRCommand) -> Choice | None:
    if command.prelude == "":
        default = 0
    else:
        try:
            default = int(t.cast(str, command.prelude))
        except ValueError:
            raise ParseException(
                f"Line {command.line} - Choice preludes must be an integer. Got:\n{command.prelude}"
            )
    if not command.children:
        raise ParseException(
            f"Line {command.line} - Choice commands need at least one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return Choice(default, *children)


def create_optional_node(command: RRCommand) -> Choice | None:
    if command.prelude not in (None, "", "skip"):
        raise ParseException(
            f"Line {command.line} - Optional preludes must be nothing or 'skip'. Got:\n{command.prelude}"
        )
    if len(command.children) != 1:
        raise ParseException(
            f"Line {command.line} - Optional commands need exactly one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return optional(children[0], skip=(command.prelude == "skip"))


def create_alternating_sequence_node(command: RRCommand) -> AlternatingSequence | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - AlternatingSequence commands cannot have preludes."
        )
    if len(command.children) != 2:
        raise ParseException(
            f"Line {command.line} - AlternatingSequence commands need exactly two children."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return AlternatingSequence(children[0], children[1])


def create_group_node(command: RRCommand) -> Group | None:
    if len(command.children) != 1:
        raise ParseException(
            f"Line {command.line} - Group commands need exactly one child."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    if command.prelude:
        return Group(children[0], label=command.prelude)
    return Group(children[0])


def create_one_or_more_node(command: RRCommand) -> OneOrMore | None:
    if command.prelude:
        raise ParseException(
            f"Line {command.line} - OneOrMore commands cannot have preludes."
        )
    if not (1 <= len(command.children) <= 2):
        raise ParseException(
            f"Line {command.line} - OneOrMore commands must have one or two children."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    return OneOrMore(*children)


def create_zero_or_more_node(command: RRCommand) -> Choice | None:
    if command.prelude not in (None, "", "skip"):
        raise ParseException(
            f"Line {command.line} - ZeroOrMore preludes must be nothing or 'skip'. Got:\n{command.prelude}"
        )
    if not (1 <= len(command.children) <= 2):
        raise ParseException(
            f"Line {command.line} - ZeroOrMore commands must have one or two children."
        )
    children = [
        _f for _f in [create_diagram(child) for child in command.children] if _f
    ]
    if not children:
        raise ParseException(
            f"Line {command.line} - ZeroOrMore has no valid children."
        )  # Unreachable?
    repeat = children[1] if len(children) == 2 else None
    return zero_or_more(children[0], repeat=repeat, skip=(command.prelude == "skip"))


def create_diagram(command: RRCommand, diagram_type="simple") -> DiagramItem | None:
    """
    From a tree of commands,
    create an actual Diagram class.
    Each command must be {command, prelude, children}
    """
    match command.name:
        case "Diagram":
            return create_diagram_node(command, diagram_type)
        case "T" | "Terminal":
            return create_terminal_node(command)
        case "N" | "NonTerminal":
            return create_non_terminal_node(command)
        case "C" | "Comment":
            return create_comment_node(command)
        case "S" | "Skip":
            return create_skip_node(command)
        case "And" | "Seq" | "Sequence":
            return create_sequence_node(command)
        case "Stack":
            return create_stack_node(command)
        case "HorizontalChoice":
            return create_horizontal_choice_node(command)
        case "MultipleChoice":
            return create_multiple_choice_node(command)
        case "OptionalSequence":
            return create_optional_sequence_node(command)
        case "Or" | "Choice":
            return create_choice_node(command)
        case "Opt" | "Optional":
            return create_optional_node(command)
        case "AlternatingSequence":
            return create_alternating_sequence_node(command)
        case "Group":
            return create_group_node(command)
        case "Plus" | "OneOrMore":
            return create_one_or_more_node(command)
        case "Star" | "ZeroOrMore":
            return create_zero_or_more_node(command)
        case _:
            raise ParseException(
                f"Line {command.line} - Unknown command '{command.name}'."
            )



# Display constants
DIAGRAM_CLASS = "railroad-diagram"  # class to put on the root <svg>
DEBUG = False  # if True, add debug info to the diagram
STROKE_ODD_PIXEL_LENGTH = (
    True  # is the stroke width an odd (1px, 3px, etc) pixel length?
)
VS = 8  # minimum vertical separation between things. For a 3px stroke, must be at least 4
AR = 10  # radius of arcs
CHAR_WIDTH = 8  # width of each monospace character. play until you find the right value for your font
COMMENT_CHAR_WIDTH = 7  # comments are in smaller text by default
INTERNAL_ALIGNMENT = (
    "center"  # how to align items when they have extra space. left/right/center
)
